import * as assert from 'assert';
import { EditorState, Change, Transaction } from 'codemirror-6';
import { type as json0 } from 'ot-json0';
import { transactionToOps, opsToTransaction } from '../src/index';

// Removes meta.time, which is the only thing that doesn't match.
const withoutTimestamp = (transaction: Transaction) => {
  // Hack around TypeScript's complaint that meta is private.
  const object = JSON.parse(JSON.stringify(transaction));
  delete object.meta.time;
  return object;
};

// Verifies that ops and transaction match in terms of behavior,
// and that the translation between ops and transaction holds in both directions.
export const verify = ({ before, after, txn, ops, txnFromOps = undefined}) => {
  const path = [];
  const state = EditorState.create({ doc: before });
  const transaction = txn(state.transaction);
  const expectedTransaction = txnFromOps ? txnFromOps(state.transaction) : transaction;

  it('applied ops should match expected text', () => {
    assert.deepEqual(after, json0.apply(before, ops));
  });

  it('inverted applied ops should match expected text', () => {
    assert.deepEqual(before, json0.apply(after, json0.invert(ops)));
  });

  it('applied transaction should match expected text', () => {
    assert.deepEqual(after, transaction.doc.toString());
  });

  it('transactionToOps', () => {
    assert.deepEqual(transactionToOps(path, transaction), ops);
  });

  it('opsToTransaction', () => {
    assert.deepEqual(
      withoutTimestamp(opsToTransaction(path, state, ops)),
      withoutTimestamp(expectedTransaction)
    );
  });
};

describe('translation (transactionToOps and opsToTransaction)', () => {
  describe('string insert', () => {
    describe('single character insert from position 0', () => {
      verify({
        before: '',
        after:'d',
        txn: transaction => transaction.change(new Change(0, 0, ['d'])),
        ops: [{ p: [0], si: 'd' }]
      });
    });
    describe('single character insert mid-string', () => {
      verify({
        before: 'HelloWorld',
        after: 'Hello World',
        txn: transaction => transaction.change(new Change(5, 5, [' '])),
        ops: [{ p: [5], si: ' ' }]
      });
    });
    describe('multiple character insert mid-string', () => {
      verify({
        before: 'HelloWorld',
        after: 'Hello Beautiful World',
        txn: transaction => transaction.change(new Change(5, 5, [' Beautiful '])),
        ops: [{ p: [5], si: ' Beautiful ' }]
      });
    });
    describe('multiple character insert at last position', () => {
      verify({
        before: 'HelloWorld',
        after: 'HelloWorldPeace',
        txn: transaction => transaction.change(new Change(10, 10, ['Peace'])),
        ops: [{ p: [10], si: 'Peace' }]
      });
    });
  });
  describe('string delete', () => {
    describe('single character delete mid-string', () => {
      verify({
        before: 'Hello World',
        after: 'HelloWorld',
        txn: transaction => transaction.change(new Change(5, 6, [''])),
        ops: [{'p': [5], 'sd':' '}]
      });
    });
    describe('multiple character delete mid-string', () => {
      verify({
        before: 'Hello Beautiful World',
        after: 'HelloWorld',
        txn: transaction => transaction.change(new Change(5, 16, [''])),
        ops: [{ p: [5], sd: ' Beautiful ' }]
      });
    });
  });
  describe('string delete and insert (replacement)', () => {
    describe('single character replacement mid-string', () => {
      verify({
        before: 'Hello World',
        after: 'Hello-World',
        txn: transaction => transaction.change(new Change(5, 6, ['-'])),
        ops: [
          {'p': [5], 'sd': ' '},
          {'p': [5], 'si': '-'}
        ],
        txnFromOps: transaction => transaction
          .change(new Change(5, 6, ['']))
          .change(new Change(5, 5, ['-']))
      });
    });
  });
  describe('newlines', () => {
    describe('newline insert', () => {
      verify({
        before: '',
        after: '\n',
        txn: transaction => transaction.change(new Change(0, 0, ['', ''])),
        ops: [{ p: [0], si: '\n' }]
      });
    });
    describe('newline delete', () => {
      verify({
        before: '\n',
        after: '',
        txn: transaction => transaction.change(new Change(0, 1, [''])),
        ops: [{ p: [0], sd: '\n' }]
      });
    });
    describe('replace with newlines', () => {
      verify({
        before: 'eat\na\npie',
        after: 'eat\nan\napple\npie',
        txn: transaction => transaction.change(new Change(4, 5, ['an', 'apple' ])),
        ops: [ { p: [4], sd: 'a' }, { p: [4], si: 'an\napple' } ],
        txnFromOps: transaction => transaction
          .change(new Change(4, 5, ['']))
          .change(new Change(4, 4, ['an', 'apple']))
      });
    });
  });
});
